{% load staticfiles %}
{% load i18n %}

<!--
Jinja2 partial to render an image

{ % include 'sharedapp/partials/full_width_image.html' with usecase=usecase spec=spec widget=widget % }

-->
<div class="block-widget-box width-all" id="widget_{{widget.id}}">
  <div class="block-filter-container d-flex flex-wrap align-items-center">
    <div class="crc-head-inside d-flex align-items-center">
      <div class="crc-head-icon"  data-bs-toggle="collapse" data-bs-target="#collapseresults_{{widget.id}}" aria-expanded="true">
        <img class="expand-open" src={% static 'gui2/appstore2/images/new-caret-icon-1.svg' %} alt="" />
        <img class="expand-add" src={% static 'gui2/appstore2/images/expand-add.svg' %} alt="" />
      </div>
      <div class="accor-right">
        <div class="block-filter-left">
          <ul class="cp-text-list d-flex align-items-center">
            <li><b>{% translate widget.name|safe %}</b></li>
          </ul>
            <p>{% translate widget.description|safe %}</p>
        </div>
      </div>
    </div>
    <div class="block-filter-option ms-auto d-flex align-items-center justify-content-center align-block-filter-option">
    </div>
  </div>
  <div id="collapseresults_{{widget.id}}" class="crc-body accordion-collapse collapse show">
    <div class="cp-body">
      <img src="{{widget.src}}" style="{{widget.style}}"></img>
    </div>
  </div>
</div>
