{% load staticfiles %}
{% load i18n %}

<!--
Jinja2 partial to render an image

{ % include 'sharedapp/partials/full_width_text.html' with usecase=usecase spec=spec widget=widget % }

-->
<div class="block-widget-box width-all" id="{{widget.id}}">
    <div class="block-filter-container d-flex flex-wrap align-items-center">
    <div class="d-flex align-items-center">
        <div class="block-filter-left">
            <ul style="font-size: 15px; font-weight: 400 !important">
                <li><b>{% translate widget.name|safe %}</b></li>
            </ul>
            <p>{% translate widget.description|safe %}</p>
        </div>
    </div>
  </div>

    <div class="market-research-content-body text-widget-font">
    {% if widget.text %}
      <p>{{widget.text|safe}}</p>
    {% else %}
             {% for k,values in widget.body_components.items %}
                {% if k != 'notes' %}
                {% endif %}
                 {% for v in values %}
                        {% include 'sharedapp/components/'|add:v.template|add:'.html' with content=v %}
                 {% endfor %}
        {% endfor %}

     {% endif %}
    </div>
{#  </div>#}
</div>
