<div class="form-group d-flex">
    <button name="submit" id="submit" value="select" type="submit" class="btn btn-default" data-toggle="tooltip" data-placement="bottom" title="Submit"><img src={% static 'gui2/appstore2/images/'|add:icon|add:'.svg' %}Select</button>
</div>