{% load static %}

<div class="cp-download" id="{{ content.title }}">
    <img {% if 'id' in content %} id="{{ content.id }}" {% else %} id="{{ content.icon }}" {% endif %} src={% static 'gui2/appstore2/images/'|add:content.icon|add:'.svg' %} alt="" />
</div>
