<!--This shows table of key value pairs-->
<div class="col-sm-{{width}} col-xs-{{width}} text-left">
 <button data-toggle="collapse" data-target="#table{{id}}"><h4>{{title}}</h4></button>
 <div id="table{{id}}" class="collapse">
   <hr/>
   <table class="table table-striped table-responsive">
     {% for p in pairs %} 
     <tr> 
       <th>{{p.name}}</th><td>{{p.value}}</td>
     </tr>
     {% endfor %} 
   </table>
 </div>
</div>
