# event name definitions
_EVNT_READY = "ready"
_EVNT_DATA = "source-loaded"
_EVNT_INDEXCREATE = "index-created"
_EVNT_INDEXLOAD = "index-loaded"
_EVNT_INDEXSAVE = "index-persisted"
_EVNT_INDEXADD = "index-doc-added"
_EVNT_QUERY = "query"
_EVNT_SANITIZE = "sanitize"
_EVNT_SPEC = "spec-loaded"
_EVNT_DOCGEN = "doc-gen"
_EVNT_STATECLEAR = "state-clear"
_EVNT_STATESTORE = "state-store"
_EVNT_STATERESTORE = "state-restore"
