{% extends 'gui2/base.html' %}
{% load dashboard_tags %}
{% load crispy_forms_tags %}
{% load humanize %}
{% load fontawesome %}
{% load control %}
{% load static %}

<!-- ============================= Address Bar Start =============================== -->
{% block breadcrumbs %}
<div id="address_content" style="display: none">
    <ol class="breadcrumb" aria-label="breadcrumb">
        <li class="breadcrumb-item"><a href="{% url 'dashboard:home' %}">Enrich</a></li>
        <li class="breadcrumb-item"><a href="{% url 'dashboard:index' %}">Usecases</a></li>
        <li class="breadcrumb-item"><a href="{% url 'dashboard:usecase_detail' usecase.org.name %}">{{usecase.org.name}}</a></li>
        <li class="breadcrumb-item active">{{spec.name}}</li>
    </ol>
</div>

{% with tags=spec.tags|join:" " %}
{% with knowledge=spec.name|add:' '|add:app.name|add:' '|add:tags %}{{ block.super }}{% endwith %}
{% endwith %}

{% endblock breadcrumbs %}
<!-- ============================= Address Bar End =============================== -->
{% block nav_index%}
<!--Header tab name-->
    {% with nav_data="apps" %}{{ block.super }}{% endwith %}
{% endblock %}
<!--============================= Help Content =============================== -->
{% block help %}
{% include 'sharedapp/partials/help.html' with app=app spec=spec %}
{{block.super}}
{% endblock %}

<!--============================= Help Content End =============================== -->
{% block content %}

<div class="container mb-30">
  <div class="row">
    <div class="col-md-12">
      <div class="row align-items-center">
        <div class="col-md-6">
          <div class="ptaa-box">
            <div class="ptaa-icon">{% fontawesome_icon app.icon %}</div>
            <div class="ptaa-content">
              <h1>{{app.verbose_name}}</h1>
              <p>{{spec.name}}: {{spec.description}}</p>
            </div>
          </div>
        </div>
        <div class="col-md-6">
          <div class="d-flex justify-content-end">
          </div>
        </div>
      </div>
    </div>
  </div>
  {% include 'dashboard/_appusers.html' with usecase=usecase spec=spec %}
  <hr>
  {% include helper with app=app spec=spec data=data %}

</div>

{% endblock %}
